//*****************************************************************************
//   Copyright (c) 1998 INRIA.
//   http://www.inria.fr/croap/aioli/modules/figue/web/credits.html#COPYRIGHT
//*****************************************************************************

package figue.internal;

import figue.resource.ColorInterface;
import figue.resource.FontInterface;
import figue.resource.ColorDefered;
import figue.resource.FontDefered;
import figue.resource.FontInitializer;
import figue.resource.BackgroundInitializer;
import figue.resource.ColorInitializer;
import figue.resource.FontMetricInterface;
import figue.resource.ResourceInitializer;
import figue.resource.FontInitializerInterface;
import figue.resource.ResourceOwnerInterface;
import figue.selection.SelectionInterface;
import figue.geometry.PositionToFather;
import figue.geometry.PositionToBrother;
import figue.geometry.Position;
import figue.geometry.Point;
import figue.format.DistanceInterface;
import figue.format.LazyDistance;
import figue.format.DecalageInterface;
import figue.format.LazyDecalage;
import figue.format.LazyDecalageToBrother;
import figue.format.Modification;
import figue.format.ModificationStatus;
import figue.error.ErrorHandler;
import figue.error.ErrorHandlerInterface;
import figue.error.Warning;
import figue.error.Error;
import figue.error.ShouldNotBeCalledError;
import figue.multithread.ThreadCondition;
import figue.multithread.ThreadConditionInterface;
import figue.box.GlyphInterface;
import figue.box.LightweightComponentAdapter;
import figue.FacadeInterface;
import figue.Constants;
import figue.Priority;

import java.util.Hashtable;

public abstract class Context extends ErrorHandler implements ContextInterface {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/internal/Context.java,v $ $revision$ $Date: 1998/11/02 15:55:12 $ Copyright 1998 INRIA.";

    protected Context(FacadeInterface aFacade,
		      String aFontName,
		      String aColorName,
		      String aBackgroundName) {
	super();
	ColorInterface theColor;
	ColorInterface theBackground;
	FontInterface theFont;
	if ( aFontName == null ) {
	    aFontName = Constants.DEFAULT_FONT;
	}
	if ( aColorName == null ) {
	    aColorName = Constants.DEFAULT_COLOR;
	}
	theColor = new ColorDefered(aColorName,new ColorInitializer(this));
	if ( aBackgroundName == null ) {
	    aBackgroundName = Constants.DEFAULT_BACKGROUND;
	}
	theBackground = new ColorDefered(aBackgroundName,new BackgroundInitializer(this));
	_facade = aFacade;
	_initialFont = new FontDefered(aFontName,new InitialFontInitializer(this));
	_initialFontMetric = null;
 	_font = new FontDefered(aFontName,new FontInitializer(this));
	_color = theColor;
	_background = theBackground;
	_colorPriority = Priority.MIN.getEnum();
	_backgroundPriority = Priority.MIN.getEnum();
	_mementoPool = new StackAsync(20,10);
	_modificationPool = new StackAsync(20,10);
	_pointPool = new StackAsync(20,10);
	_positionToFatherPool = new StackAsync(20,10);
	_positionToBrotherPool = new StackAsync(20,10);
	_lazyDistancePool = new StackAsync(20,10);
	_lazyDecalagePool = new StackAsync(20,10);
	_lazyDecalageToBrotherPool = new StackAsync(20,10);
	_pairGlyphPathPool = new StackAsync(20,10);
	_flyWeightRegistryForString = new Hashtable(Globals.AVERAGE_FLY_WEIGHT_ATOM_NUMBER,
						    Globals.FAST_ACCESS);
	_flyWeightRegistryForKeyWord = new Hashtable(Globals.AVERAGE_FLY_WEIGHT_ATOM_NUMBER,
						     Globals.FAST_ACCESS);
	final int theMin = '\u0000';
	final int theMax = '\uFFFF';
	_flyWeightRegistryForChar = new Hashtable[theMax - theMin + 1]; /// WASTED SPACE
	_formatAdvance = new ThreadCondition("Format advance (context)");
	_majorFormatAdvance = new ThreadCondition("Major format advance (context)");
	_buildAdvance = new ThreadCondition("Build advance (context)");
    }
  
    public final FacadeInterface getFacade() {
	return _facade;
    }

    public final FontInterface getFont() {
	return _font;
    }

    public void setFont(FontInterface aFont) {
	_font = aFont;
    }

    public final void setInitialFont(FontInterface aFont) {
	_initialFont = aFont;
	_font = aFont;
    }

    public final FontInterface getInitialFont() {
	return _initialFont;
    }

    public final ColorInterface getColor() {
	return _color;
    }

    public void setColor(ColorInterface aColor) {
	_color = aColor;
    }

    public final ColorInterface getBackground() {
	return _background;
    }

    public final void setBackground(ColorInterface aColor) {
	_background = aColor;
    }

    public final synchronized FontMetricInterface getFontMetric() {
	if ( _initialFontMetric == null ) {
	    _initialFont.getKey(); /// BEURK
	    _initialFontMetric = getFontMetric(_initialFont);
	}
	return _initialFontMetric;
    }

    public final ContextMementoInterface set(FontInterface aFont,
					     ColorInterface aColor, 
					     ColorInterface aBackground) {
	final int thePriority = Priority.MIN.getEnum();
	FontInterface theOldFont = null;
	ColorInterface theOldColor = null; 
	ColorInterface theOldBackground = null;
	int theOldColorPriority = -1;
	int theOldBackgroundPriority = -1;
	if ( ( aFont != null ) && ( aFont != _font ) ) {
	    theOldFont = _font;
	    setFont(aFont);
	}
	if ( ( aColor != null ) && ( thePriority >= _colorPriority ) ) {
	    theOldColorPriority = _colorPriority;
	    _colorPriority = thePriority;
	    theOldColor = _color;
	    setColor(aColor);
	}
	if ( ( aBackground != null ) && ( thePriority >= _backgroundPriority ) ) {
	    theOldBackgroundPriority = _backgroundPriority;
	    _backgroundPriority = thePriority;
	    theOldBackground = _background;
	    setBackground(aBackground);
	}
	return getTrashMemento(theOldFont,
			       theOldColor,
			       theOldBackground,
			       theOldColorPriority,
			       theOldBackgroundPriority);
    }

    public final ContextMementoInterface set(SelectionInterface aSelection) {
	final int thePriority = aSelection.getPriority();
	ColorInterface theColor = aSelection.getColor();
	ColorInterface theBackground = aSelection.getBackground();
	FontInterface theOldFont = null;
	ColorInterface theOldColor = null; 
	ColorInterface theOldBackground = null;
	int theOldColorPriority = -1;
	int theOldBackgroundPriority = -1;
	if ( ( theColor != null ) && ( thePriority > _colorPriority ) ) {
	    theOldColorPriority = _colorPriority;
	    _colorPriority = thePriority;
	    theOldColor = _color;
	    setColor(theColor);
	}
	if ( ( theBackground != null ) && ( thePriority > _backgroundPriority ) ) {
	    theOldBackgroundPriority = _backgroundPriority;
	    _backgroundPriority = thePriority;
	    theOldBackground = _background;
	    setBackground(theBackground);
	}
	return getTrashMemento(theOldFont,
			       theOldColor,
			       theOldBackground,
			       theOldColorPriority,
			       theOldBackgroundPriority);
    }

    public final void set(SelectionInterface aSelection,
			  ContextMementoInterface aMemento) {
	final int thePriority = aSelection.getPriority();
	ColorInterface theColor = aSelection.getColor();
	ColorInterface theBackground = aSelection.getBackground();
	FontInterface theOldFont = null;
	ColorInterface theOldColor = null; 
	ColorInterface theOldBackground = null;
	int theOldColorPriority = -1;
	int theOldBackgroundPriority = -1;
	if ( ( theColor != null ) && 
	     ( thePriority > _colorPriority ) && 
	     ( aMemento.getColor() == null ) ) {
	    aMemento.setColor(_color);
	    aMemento.setColorPriority(_colorPriority);
	    _colorPriority = thePriority;
	    setColor(theColor);
	}
	if ( ( theBackground != null ) && 
	     ( thePriority > _backgroundPriority ) && 
	     ( aMemento.getBackground() == null ) ) {
	    aMemento.setBackground(_background);
	    aMemento.setBackgroundPriority(_backgroundPriority);
	    _backgroundPriority = thePriority;
	    setBackground(theBackground);
	}
    }

    public final void setState(ContextMementoInterface aMemento) {
	if ( aMemento.getFont() != null ) {
	    setFont(aMemento.getFont());
	}
	if ( aMemento.getColor() != null ) {
	    _colorPriority = aMemento.getColorPriority();
	    setColor(aMemento.getColor());
	}
	if ( aMemento.getBackground() != null ) {
	    _backgroundPriority = aMemento.getBackgroundPriority();
	    setBackground(aMemento.getBackground());
	}
    }

    public final void unapplyState(ContextMementoInterface aMemento) { 
	if ( aMemento.getFont() != null ) {
	    _font = aMemento.getFont();
	}
    }

    public final ContextMementoInterface saveState() {
	return getTrashMemento(_font,_color,_background,_colorPriority,_backgroundPriority);
    }

    public final ContextMementoInterface apply(FontInterface aFont) { 
	final int thePriority = Priority.MIN.getEnum();
	FontInterface theOldFont = null;
	if ( ( aFont != null ) && ( aFont != _font ) ) {
	    theOldFont = _font;
	    _font = aFont;
	}
	return getTrashMemento(theOldFont,null,null,-1,-1);
    }

    public boolean isComplientWith(LightweightComponentAdapter aLightweight) {
	return true;
    }

    private final ContextMementoInterface getTrashMemento(FontInterface aFont,
							  ColorInterface aColor, 
							  ColorInterface aBackground,
							  int aColorPriority,
							  int aBackgroundPriority) {
	if ( ! _mementoPool.isEmpty() ) {
	    ContextMementoInterface theMemento = (ContextMementoInterface)_mementoPool.pop();
	    theMemento.set(aFont,aColor,aBackground,aColorPriority,aBackgroundPriority);
	    return theMemento;
	}
	return new ContextMemento(aFont,aColor,aBackground,aColorPriority,aBackgroundPriority);
    }
  
    public final void disposeMemento(ContextMementoInterface aMemento) {
	_mementoPool.push(aMemento);
    }

    public final void disposeModification(Modification aModification) {
	_modificationPool.push(aModification);
    }

    public final Modification newModification(ModificationStatus aStatus) {
	if ( ! _modificationPool.isEmpty() ) {
	    final Modification theModification = (Modification)(_modificationPool.pop());
	    theModification.init(aStatus);
	    return theModification;
	}
	else {
	    return new Modification(aStatus);
	}
    }
    
    public final Modification newModification() {
	return newModification(ModificationStatus.NONE);
    }

    public final void disposeLazyDecalage(LazyDecalage aDecalage) {
	_lazyDecalagePool.push(aDecalage);
    }
    
    public final LazyDecalage newLazyDecalage(GlyphInterface aFather) {
	return newLazyDecalage(aFather,null);
    }

    public final LazyDecalage newLazyDecalage(GlyphInterface aFather,
					      DecalageInterface aDecalage) {
	if ( ! _lazyDecalagePool.isEmpty() ) {
	    LazyDecalage theDecalage = (LazyDecalage)(_lazyDecalagePool.pop());
	    theDecalage.init(aFather,aDecalage);
	    return theDecalage;
	}
	else {
	    return new LazyDecalage(aFather,aDecalage);
	}
    }
    
    public final void disposeLazyDecalageToBrother(LazyDecalageToBrother aDecalage) {
	_lazyDecalageToBrotherPool.push(aDecalage);	
    }
    
    public final LazyDecalageToBrother newLazyDecalageToBrother(GlyphInterface aFather,
								GlyphInterface aBrother,
								DecalageInterface aDecalage) {
	if ( ! _lazyDecalageToBrotherPool.isEmpty() ) {
	    LazyDecalageToBrother theDecalage = (LazyDecalageToBrother)(_lazyDecalageToBrotherPool.pop());
	    theDecalage.init(aFather,aBrother,aDecalage);
	    return theDecalage;
	}
	else {
	    return new LazyDecalageToBrother(aFather,aBrother,aDecalage);
	}
    }
    
    public final void disposeLazyDistance(LazyDistance aDistance) {
	_lazyDistancePool.push(aDistance);
    }
    
    public final LazyDistance newLazyDistance(GlyphInterface aGlyph,
					      DistanceInterface aDistance,
					      DecalageInterface aDecalage) {
	if ( ! _lazyDistancePool.isEmpty() ) {
	    LazyDistance theDistance = (LazyDistance)(_lazyDistancePool.pop());
	    theDistance.init(aGlyph,aDistance,aDecalage);
	    return theDistance;
	}
	else {
	    return new LazyDistance(aGlyph,aDistance,aDecalage);
	}
    }

    public final void disposePairGlyphPath(PairGlyphPath aPair) {
	_pairGlyphPathPool.push(aPair);
    }

    public PairGlyphPath newPairGlyphPath(GlyphInterface aGlyph) {
	if ( ! _pairGlyphPathPool.isEmpty() ) {
	    PairGlyphPath thePair = (PairGlyphPath)(_pairGlyphPathPool.pop());
	    thePair.setGlyph(aGlyph);
	    thePair.setPath(null);
	    return thePair;
	}
	return new PairGlyphPath(aGlyph);
    }

    public final Point newPoint() {
	if ( ! _pointPool.isEmpty() ) {
	    return (Point)_pointPool.pop();
	}
	return new Point();
    }
  
    public final void disposePoint(Point aPoint) {
	_pointPool.push(aPoint);
    }

    public final PositionToFather newPositionToFather() {
	if ( ! _positionToFatherPool.isEmpty() ) {
	    return (PositionToFather)_positionToFatherPool.pop();
	}
	return new PositionToFather();
    }
  
    public final void disposePositionToFather(PositionToFather aPosition) {
	_positionToFatherPool.push(aPosition);
    }
  
    public final PositionToBrother newPositionToBrother(Position aPosition) {
	if ( ! _positionToBrotherPool.isEmpty() ) {
	    PositionToBrother thePosition 
		= (PositionToBrother)_positionToBrotherPool.pop();
	    thePosition.setLocation(aPosition);
	    return thePosition;
	}
	return new PositionToBrother(aPosition);
    }
  
    public final void disposePositionToBrother(PositionToBrother aPosition) {
	_positionToBrotherPool.push(aPosition);
    }

    public final Hashtable[] getFlyWeightRegistryForChar() {
	return _flyWeightRegistryForChar;
    }
    public final Hashtable getFlyWeigthRegistryForString() {
	return _flyWeightRegistryForString;
    }
    public Hashtable getFlyWeigthRegistryForKeyWord() {
	return _flyWeightRegistryForKeyWord;
    }

    public final void awaitMajorFormatAdvance() {
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("-> awaitMajorFormatAdvance");
	}
	synchronized ( _majorFormatAdvance ) {
	    _majorFormatAdvance.await(); 
	}
	if ( Globals.DEBUG_MULTI_THREAD ) {
	    System.out.println("<- awaitMajorFormatAdvance");
	}
    }

    public final void awaitStable(GlyphInterface aGlyph) {
	if ( ! aGlyph.isUnstable() ) {
	    return;
	}
	else {
	    synchronized ( _buildAdvance ) {
		while ( true ) {
		    synchronized ( aGlyph ) {
			if ( ! aGlyph.isUnstable() ) {
			    break;
			}
		    }
		    _buildAdvance.await();
		}
	    }
	}
    }

    public final void awaitClean(GlyphInterface aGlyph) {
	if ( ! aGlyph.isClean() ) {
	    awaitStable(aGlyph);
	    synchronized ( _formatAdvance ) {
		while ( true ) {
		    synchronized ( aGlyph ) {
			if ( aGlyph.isClean() ) {
			    break;
			}
		    }
		    _formatAdvance.await();
		}
	    }
	}
    }

    public final void formatAdvance() {
	_formatAdvance.signal();
    }

    public void majorFormatAdvance() {
	_majorFormatAdvance.signal();
	_formatAdvance.signal();
    }

    public void lastMajorFormatAdvance() { /// +++
	_majorFormatAdvance.signal();
	_formatAdvance.signal();
    }

    public void lastMajorFormatInteractiveAdvance() { /// +++ 
	_majorFormatAdvance.signal();
	_formatAdvance.signal();
    }

    public final void buildAdvance() {
	_buildAdvance.signal();
    }

    public final void buildComplete() {
	_buildAdvance.signal();
    }

    public final boolean hasNextErrorHandler() {
	return ( getFacade() != null );
    }
  
    public final ErrorHandlerInterface getNextErrorHandler() {
	return getFacade();
    }

    public final void manageFatalError(Error anError) {
	throw new ShouldNotBeCalledError();
    }
  
    public final void manageWarning(Warning aWarning) {
	throw new ShouldNotBeCalledError();
    }

    public final boolean canHandleError() {
	return false;
    }

    public void dispose() {
	_facade = null;
	_initialFont = null;
	_initialFontMetric = null;
	_font = null;
	_color = null;
	_background = null;
	_mementoPool = null;
	_modificationPool = null;
	_pointPool = null;
	_positionToFatherPool = null;
	_positionToBrotherPool = null;
	_lazyDistancePool = null;
	_lazyDecalagePool = null;
	_lazyDecalageToBrotherPool = null;
	_pairGlyphPathPool = null;
	_flyWeightRegistryForString.clear();
	_flyWeightRegistryForString = null;
	_flyWeightRegistryForChar = null;
	_formatAdvance = null;
	_majorFormatAdvance = null;
	_buildAdvance = null;
    }

    public String toString() {
	StringBuffer theResult = new StringBuffer("Context<");
	theResult.append(_font.toString()).append(", ")
	    .append(_color.toString()).append("/").append(_background.toString())
	    .append(">");
	return theResult.toString();
    }

    private FacadeInterface _facade;
    private FontInterface _initialFont;
    private FontMetricInterface _initialFontMetric;
    private FontInterface _font;
    private ColorInterface _color;
    private ColorInterface _background;
    private int _colorPriority;
    private int _backgroundPriority;
    private StackInterface _mementoPool;
    private StackInterface _modificationPool;
    private StackInterface _pointPool;
    private StackInterface _positionToFatherPool;
    private StackInterface _positionToBrotherPool;
    private StackInterface _lazyDistancePool;
    private StackInterface _lazyDecalagePool;
    private StackInterface _lazyDecalageToBrotherPool;
    private StackInterface _pairGlyphPathPool;
    private Hashtable _flyWeightRegistryForString;
    private Hashtable _flyWeightRegistryForKeyWord;
    private Hashtable[] _flyWeightRegistryForChar;
    private ThreadConditionInterface _formatAdvance;
    private ThreadConditionInterface _majorFormatAdvance;
    private ThreadConditionInterface _buildAdvance;
  
}

final class ContextMemento implements ContextMementoInterface {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/internal/Context.java,v $ $revision$ $Date: 1998/11/02 15:55:12 $ Copyright 1998 INRIA.";

    public ContextMemento(FontInterface aFont,
			  ColorInterface aColor, 
			  ColorInterface aBackground,
			  int aColorPriority,
			  int aBackgroundPriority) {
	_font = aFont;
	_color = aColor;
	_background = aBackground;
	_colorPriority = aColorPriority;
	_backgroundPriority = aBackgroundPriority;
    }

    public final void set(FontInterface aFont,
			  ColorInterface aColor, 
			  ColorInterface aBackground,
			  int aColorPriority,
			  int aBackgroundPriority) {
	_font = aFont;
	_color = aColor;
	_background = aBackground;
	_colorPriority = aColorPriority;
	_backgroundPriority = aBackgroundPriority;
    }

    public final FontInterface getFont() {
	return _font;
    }

    public final ColorInterface getColor() {
	return _color;
    }
  
    public final void setColor(ColorInterface aColor) {
	_color = aColor;
    }
  
    public final ColorInterface getBackground() {
	return _background;
    }
  
    public final void setBackground(ColorInterface aBackground) {
	_background = aBackground;
    }

    public final int getColorPriority() {
	return _colorPriority;
    }
  
    public final void setColorPriority(int aPriority) {
	_colorPriority = aPriority;
    }

    public final int getBackgroundPriority() {
	return _backgroundPriority;
    }
  
    public final void setBackgroundPriority(int aPriority) {
	_backgroundPriority = aPriority;
    }

    public final boolean isFull() {
	return ( ( _color != null ) && 
		 ( _background != null ) );
    }

    private FontInterface _font;
    private ColorInterface _color;
    private ColorInterface _background;
    private int _colorPriority;
    private int _backgroundPriority;

}

final class InitialFontInitializer extends ResourceInitializer implements FontInitializerInterface {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/internal/Context.java,v $ $revision$ $Date: 1998/11/02 15:55:12 $ Copyright 1998 INRIA.";
  
    protected InitialFontInitializer(ContextInterface anOwner) {
	super();
	_owner = anOwner;
    }

    public final ResourceOwnerInterface getOwner() {
	return _owner;
    }

    public final FontInterface initResource(String aName) {
	FontInterface theFont = newFont(_owner,aName);
	_owner.setInitialFont(theFont);
	return theFont;
    }

    public final String getName() {
	return "Font";
    }
  
    private ContextInterface _owner;
  
}
