//*****************************************************************************
//   Copyright (c) 1998 Bernard Sufrin, Oxford.
//   http://www.comlab.ox.ac.uk/oucl/people/bernard.sufrin.html
//*****************************************************************************

package figue.jape;

import  java.awt.*;
import  java.awt.image.*;
import  java.util.*;
import  java.net.URL;
import  java.io.Serializable;

/**
        A <TT>TypeFace</TT> whose glyphs are
        represented as <TT>ImageChar</TT>s.
        
        @see ImageChar
*/

public class ImageFont implements Font, ColorScheme, ColorModelGen {  

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/jape/ImageFont.java,v $ $revision$ $Date: 1998/11/02 15:55:23 $ Copyright 1998, Bernard Sufrin, Oxford.";

    /** Character images. */
    private ImageChar[] image = new ImageChar[256];

    private int   maxheight  = 0, 
	maxdescent = 0,
	scale,
	bits;

    private ColorScheme cs;
    private java.awt.image.ColorModel  cm;
   
    /** Get the appropriate color model for this font. This
	is delayed until the scaling of the first character in the
	font takes place, so that computation of the foregound
	and background colours to be used can be delayed. The first
	scaling usually takes place when a character is to be drawn
	for the first time, and this is after fore and background
	colours have been established.
	<P>
	Only 8-bit and 1-bit representations are
	handled correctly at present. The 1-bit representations are
	rather hideous at small sizes, but passable at large and huge
	sizes.
    */
   
    public  java.awt.image.ColorModel getColorModel()
    { int mscale = bits==8?scale:1;
    int mbits  = bits==8?8:1;
    if (cm==null) cm = ColorModel.range(mbits, mscale, getForeground(), getBackground());
    return cm;
    }
   
   
    private BitMapFont          base;
    private Component           component;
   
    public Color    getForeground() { return cs.getForeground(); }
    public Color    getBackground() { return cs.getBackground(); }
    public int      getScale()      { return scale; }

       
    public Font getFont(Color fg, Color bg)
    { return ImageFont.open(component, base, scale, bits, ColorModel.from(fg, bg));
    }
   
    public Font getFont(ColorScheme sc)
    { return ImageFont.open(component, base, scale, bits, sc);
    }
   
    public Font getFont()
    { return ImageFont.open(component, base, scale, bits, ColorModel.from(component));
    }
   
    /** Make an Image font suitable for display in the given component, using
	the given base bitmap font at the given linear scale. The antialiasing
	color model is generated, when needed, from the given ColorScheme.
    */
    private  ImageFont(Component component, BitMapFont base, int scale, int bits, ColorScheme cs)
    { BitMapChar defaultChar = null;
    this.scale     = scale;
    this.bits      = bits;
    this.base      = base;
    this.component = component;
    this.cs        = cs;
    makeImages();
    }
   
   
    /** Make an Image font suitable for display in the given component, using
	the given base bitmap font at the given linear scale. The antialiasing
	color model is generated, when needed, from the given ColorScheme.
	<P>
	Most public (and internal) access to image fonts is through here.
    */
    public static ImageFont open(Component component, final BitMapFont base, int scale, int bits, final ColorScheme cs)
    { Color fg = cs.getForeground();
    Color bg = cs.getBackground();
    // If we can't figure out the colour scheme, then we can't cache the font.
    // This only happens when fonts are constructed for use in components
    // whose foreground and background colours haven't yet been set
    // (because they are not mapped or because they haven't been 
    // set explicitly). 
    if (fg==null || bg==null) 
	{ 
	    //       LOG.Debug("ImageFont.open: "+base);
	    return new ImageFont(component, base, scale, bits, cs);
	}
    else
	{ String key = base+"#"+bits+"%"+fg+bg;
	ImageFont result = (ImageFont) cache.get(key);
	if (result==null) 
	    { result = new ImageFont(component, base, scale, bits, cs);
	    cache.put(key, result);
	    //         LOG.Debug("ImageFont.open: "+key);
	    }
	return result;
	}
    }
   
    private static Hashtable cache = new Hashtable();
   
    /**
       Clear the font cache. Fonts created for use with
       a component (such as a window or frame) 
       must be removed from the cache when the
       component is disposed of. At present this is the
       only way of doing so.
    */
    public static void clearCache()
    { cache = new Hashtable();
    }
   
      
    private void makeImages()
    { BitMapChar defaultChar = null;
    for (int chn=0; chn<256; chn++)
	{  BitMapChar ch = base.getChar(chn);
        if (ch!=null) 
	    { image[chn] = new ImageChar(component, ch, this);
	    if (defaultChar==null && ch.width>0 && ch.height>0) defaultChar = ch;
	    }
	}
    if (defaultChar != null)
	{  
	    setDefaults(new ImageChar(component, defaultChar.box(scale), this));           
	}
    maxheight  = base.getHeight();
    maxdescent = base.getDescent();
    // A space in the font is an en width if there isn't a space model
    if (image[' ']!=null && image[' '].width==0)
	{  if (image['n']!=null) image[' '].width=image['n'].width;
	}
    }
   
   
    /** Height of characters in the font */  
    public int getHeight()  { return maxheight; }
   
    /** Maximum ascent of characters in the font */
    public int getAscent()  { return maxheight - maxdescent; }
   
    /** Maximum descent of characters in the font */
    public int getDescent() { return maxdescent; }
   
    /** Suggested leading */
    public int getLeading() { return Math.max(1, getHeight() / 10); } 
   
    private void setDefaults(ImageChar ch)
    {
	for (int chn=0; chn<256; chn++) 
	    if (image[chn]==null) image[chn] = ch;
    }
      
    /** Return the specified glyph */
    public  final ImageChar getChar(int i)  { return image[i]; }
   
   
    /**
       Return a (monotonically-increasing) array of right-hand boundaries of
       the characters in the string.
    */
  
    public final int[] getBoundaries(String s)
    { return getBoundaries(s.toCharArray()); }
   
    /**
       Return a (monotonically-increasing) array of right-hand boundaries of
       the characters.
    */   
    public final int[] getBoundaries(char[] chars)
    { int[] bs = new int[chars.length];
    int b = 0;
    for (int i=0; i<bs.length; i++)
	{ b+=image[chars[i]].width-image[chars[i]].xoffset;
	bs[i]=b;
	}
    return bs;          
    }
   
   
    /** Draw the given <TT>char[]</TT> in the given <TT> Graphics</TT>
	at the given location.
    */
   
    public final void draw(Graphics g, char[] chars, int x, int baseline)
    { 
	for (int i=0; i<chars.length; i++)
	    {  char c = chars[i];
	    ImageChar ch    = image[c];
	    int top   = baseline - ch.yoffset,
		left  = x    - ch.xoffset,
		bot   = top  + ch.height,
		right = left + ch.width;
	    //  Space character is just a MODEL for a space
	    if  (c!=' ') g.drawImage(ch.image(), left, top, null);
	    x = right;
	    }
    }
   
    /** Draw the given <TT> String</TT> in the given <TT> Graphics</TT>
	at the given location.
    */
   
    public final void draw(Graphics g, String chars, int x, int baseline)
    { 
	for (int i=0; i<chars.length(); i++)
	    {  char c = chars.charAt(i);
	    ImageChar ch    = image[c];
	    int top   = baseline - ch.yoffset,
		left  = x    - ch.xoffset,
		bot   = top  + ch.height,
		right = left + ch.width;
	    //  Space character is just a MODEL for a space
	    if  (c!=' ') g.drawImage(ch.image(), left, top, null);
	    x = right;
	    }
    }
      
    public final int getWidth(String s)
    { return getDimension(s).width;
    }
   
    /** Dimension of the bounding box of the given character array */
    public final Dimension getDimension(char[] chars)
    { int w=0, h=0, d=0;
    for (int i=0; i<chars.length; i++)
	{ ImageChar ch=image[chars[i]];
	w+=ch.width-ch.xoffset;
	if (h<ch.height) h=ch.height;
	if (d<ch.height-ch.yoffset) d = ch.height-ch.yoffset;
	}
    return new Dimension(w, h+d);
    }
   
    /** Dimension of the bounding box of the given <TT> String</TT> */
    public final Dimension getDimension(String chars)
    { int w=0, h=0, d=0;
    for (int i=0; i<chars.length(); i++)
	{ ImageChar ch=image[chars.charAt(i)];
	w+=ch.width-ch.xoffset;
	if (h<ch.height) h=ch.height;
	if (d<ch.height-ch.yoffset) d = ch.height-ch.yoffset;
	}
    return new Dimension(w, h+d);
    }

}




























