//*****************************************************************************
//   Copyright (c) 1998 Bernard Sufrin, Oxford.
//   http://www.comlab.ox.ac.uk/oucl/people/bernard.sufrin.html
//*****************************************************************************

// Inspired by the work of Garth Dickie on IDVI, and derived
// in part from his implementation. 

package figue.jape;

import  java.io.*;
import  java.net.URL;
import  java.util.*;

public class PKBitMapFont extends SimpleBitMapFont { 

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/jape/PKBitMapFont.java,v $ $revision$ $Date: 1998/11/02 15:55:24 $ Copyright 1998, Bernard Sufrin, Oxford.";

    /** Read a PK file from a URL and construct a bitmap front from it. */
    private PKBitMapFont(URL fontbase, String name, int dpi)
	throws java.net.MalformedURLException, 
	       java.io.IOException 
    { super(name, dpi);
    URL       font = new URL(fontbase, name + "." + dpi + "pk");
    RawPKFont raw  = new RawPKFont(font.openStream(), font.toString());
    setBitMaps(raw.bitmaps);
    this.name = font.toString();
    }
  
    public String toString() { return name; }
  
    static private Hashtable cache = new Hashtable();
  
    /** Open a scaled bitmap font and cache the result */
    static public BitMapFont openScaled(URL fontbase, String name, int dpi, int scale)
	throws java.net.MalformedURLException, 
	       java.io.IOException 
    { String     key    = fontbase + name + "." + dpi + "pk" + "@" + scale;
    BitMapFont result = (BitMapFont) cache.get(key);
    if (result!=null) return result;
    result = open(fontbase,  name, dpi).scale(scale);
    cache.put(key, result);
    //    LOG.Debug("PKBitMapFont: scaled "+name + "." + dpi + "pk" + "@" + scale);
    return result;
    }
  
    /** Fetch an unscaled bitmap font and cache the result */
    static public BitMapFont open(URL fontbase, String name, int dpi)
	throws java.net.MalformedURLException, 
	       java.io.IOException 
    { String     key    = fontbase + name + "." + dpi + "pk" + "@" + 1;
    BitMapFont result = (BitMapFont) cache.get(key);
    if (result!=null) return result;
    result = new PKBitMapFont(fontbase,  name, dpi);
    cache.put(key, result);
    //    LOG.Debug("PKBitMapFont fetched "+fontbase + name + "." + dpi + "pk" );
    return result;
    }
  
    static public void clearCache() { cache = new Hashtable(); }
  
}

final class RawPKFont {   

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/jape/PKBitMapFont.java,v $ $revision$ $Date: 1998/11/02 15:55:24 $ Copyright 1998, Bernard Sufrin, Oxford.";

    protected RawPKFont( InputStream is, String sourceName)
	throws IOException 
    {
        
        this.sourceName = sourceName;

        pks = new ByteCodedInputStream( is );

        if( pks.unsigned2( ) != Magic )
            throw new IOException
		( "RawPKFont: stream not a PK file: "+sourceName);

        pks.skip( pks.unsigned1( ) );     //  comment
        pks.skip( 4 );                    //  design size
        checksum = pks.unsigned4( );      //  checksum to compare with dvi file
        pks.skip( 8 );                    //  hppp and vppp
        
        loadFont();
    }
    



    // PK commands
    final private static int    Magic   = 0xF759;
    final private static int    CmdX1   = 240;
    final private static int    CmdX2   = 241;
    final private static int    CmdX3   = 242;
    final private static int    CmdX4   = 243;
    final private static int    CmdY    = 244;
    final private static int    CmdPOST = 245;
    final private static int    CmdNOOP = 246;
    final private static int    CmdPRE  = 247;
    final private static int    Cmd248  = 248;
    final private static int    Cmd249  = 249;
    final private static int    Cmd250  = 250;
    final private static int    Cmd251  = 251;
    final private static int    Cmd252  = 252;
    final private static int    Cmd253  = 253;
    final private static int    Cmd254  = 254;
    final private static int    Cmd255  = 255;

    public  BitMapChar[ ]        bitmaps = new BitMapChar[ 256 ];
    private long                 checksum;
    private ByteCodedInputStream pks;
    private String               sourceName;

    private void loadFont() 
    {
        try 
	    {
		boolean done = false;
		while( ! done ) 
		    {
			int command = pks.unsigned1( );
			switch( command ) 
			    {
			    default:       decodeCharacter(command );    break;
			    case CmdX1:    pks.skip( pks.unsigned1( ));  break;
			    case CmdX2:    pks.skip( pks.unsigned2( ));  break;
			    case CmdX3:    pks.skip( pks.unsigned3( ));  break;
			    case CmdX4:    pks.skip( pks.unsigned4( ));  break;
			    case CmdY:     pks.skip( 4 );                break;
			    case CmdPOST:  done = true;                  break;
			    case CmdNOOP:  break;
			    case CmdPRE:
			    case Cmd248:
			    case Cmd249:
			    case Cmd250:
			    case Cmd251:
			    case Cmd252:
			    case Cmd253:
			    case Cmd254:
			    case Cmd255:
				throw new IOException
				    ("RawPKFont: bad command "+command+" in PK stream "+sourceName);
			    }
		    }
		pks.close( );
		pks = null;
	    } 
        catch
	    ( Exception e ) 
	    {
		lastError=e;
	    }
    }

    private void decodeCharacter(int flags )
	throws IOException 
    {
        int byteCount;
        int index;
        int integerSize;

        if(( flags & 0x7 ) == 0x7 ) {
            byteCount = pks.unsigned4( );
            index = pks.unsigned4( );
            integerSize = 4;
        } else if(( flags & 0x4 ) == 0x4 ) {
            byteCount = (( flags & 0x3 ) << 16 ) + pks.unsigned2( );
            index = pks.unsigned1( );
            integerSize = 2;
        } else {
            byteCount = (( flags & 0x3 ) << 8 ) + pks.unsigned1( );
            index = pks.unsigned1( );
            integerSize = 1;
        }

        int dynf = flags >> 4;
        boolean paint = ( flags & 0x8 ) != 0;

        bitmaps[index] = readPKBitMapChar( dynf, paint, integerSize, byteCount );
    }
    
    private BitMapChar readPKBitMapChar(int dynf, boolean paint, int integerSize, int byteCount )
	throws IOException 
{ 
    int tfmWidth = pks.signed( integerSize == 4 ? 4 : 3 );
    pks.skip( integerSize == 4 ? 8 : integerSize );
    int width   = pks.unsigned( integerSize );
    int height  = pks.unsigned( integerSize );
    int xOffset = pks.signed( integerSize );
    int yOffset = pks.signed( integerSize );
    int used    = 4 * integerSize + ( integerSize == 4 ? 12 : 3 + integerSize );
    byte[] encodedBits = new byte[ byteCount - used ];
    pks.read( encodedBits );
    
    // Now decode the bits
    
    byte[] bits = new byte[width*height];
    
    if (dynf == 14)
	{ // uncompressed
	    BitStream bitS = new BitStream( encodedBits );
	    for( int i = 0; i < bits.length; ++ i )
		bits[ i ] = ( byte ) bitS.bit( );
	}
    else
	{ // compressed
	    NybbleStream nybS = new NybbleStream( encodedBits );
	    int     bitIndex  = 0;
	    int     rowRepeat = 0;
	    boolean setRowRepeat = false;
	    while( bitIndex < bits.length ) 
		{   
		    int command = nybS.nybble( );
		    if ( command == 0xF ) { rowRepeat = 1; }
		    else 
			if ( command == 0xE ) { setRowRepeat = true; } 
			else 
			    {   
				int value;
				if ( command > dynf ) 
				    value = 16 * command + nybS.nybble( ) - 15 * ( dynf + 1 );
				else 
				    if ( command > 0 ) 
					value = command;
				    else 
					{
					    int count = 1;
					    while (( value = nybS.nybble( )) == 0 ) ++ count;                
					    while (( count -- ) != 0 ) value = ( value << 4 ) | nybS.nybble( );                
					    value = value + 193 - 15 * dynf;
					}

				if ( setRowRepeat ) 
				    {
					rowRepeat = value;
					setRowRepeat = false;
				    } 
				else 
				    {
					//  set bits in the array
					if ( rowRepeat != 0 ) 
					    {
						int widthLeft = width - bitIndex % width;
						if( widthLeft <= value ) {
						    value -= widthLeft;

						    if ( paint )
							for (int target = bitIndex+widthLeft; bitIndex<target; ++bitIndex )
							    bits[ bitIndex ] = 1;
						    else 
							bitIndex += widthLeft;

						    for (int target = bitIndex + rowRepeat * width; bitIndex < target; ++ bitIndex )
							bits[ bitIndex ] = bits[ bitIndex - width ];

						    rowRepeat = 0;
						}
					    }

					if ( paint )
					    for ( int target = bitIndex + value; bitIndex < target; ++ bitIndex )
						bits[ bitIndex ] = 1;
					else bitIndex += value;
					paint = ! paint;
				    }
			    }
		}        
	} // end compressed
    return new BitMapChar(width, height, xOffset, yOffset, bits);
}
    
private static Exception lastError;
public  static Exception lastError() { return lastError; }


    
}

final class NybbleStream {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/jape/PKBitMapFont.java,v $ $revision$ $Date: 1998/11/02 15:55:24 $ Copyright 1998, Bernard Sufrin, Oxford.";

    byte[ ] buffer;
    int     index = 0;

    int     currentByte;
    boolean more = false;

    NybbleStream( byte[ ] buffer ) { this.buffer = buffer; }

    int nybble( ) 
    {
        if ( more ) 
	    {
		more = false;
		return currentByte & 0xF;
	    } 
        else 
	    {
		more = true;
		currentByte = buffer[ index ++ ];
		return ( currentByte >> 4 ) & 0xF;
	    }
}
}

final class BitStream {

    /**
     * For code identification with unix what command.
     * @level internal
     */
    public final static String VERSION_ID
	= "@(#) $Source: /net/croap/CVSROOT/figue/java/jape/PKBitMapFont.java,v $ $revision$ $Date: 1998/11/02 15:55:24 $ Copyright 1998, Bernard Sufrin, Oxford.";

    byte[ ] buffer;
    int     index = 0;

    int     currentByte;
    int     shift = 0;

    BitStream( byte[ ] buffer ) { this.buffer = buffer; }

    int bit( ) 
    {
        if ( shift == 0 ) 
	    {
		currentByte = buffer[ index ++ ];
		shift = 8;
	    }
        return ( currentByte >> -- shift ) & 1;
    }
}














